#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <memory.h>
#ifndef sun
#include <fcntl.h>
#endif
#ifdef _AIX
#include <sys/select.h>
#endif

#include <sys/socket.h>
#ifdef _AIX
#include <sys/socketvar.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>

extern void log_file();
char *get_msg();
char *send_msg();
char *string_copy();
void do_exit();

extern int errno;

static	char	*left_to_mud = 0, *left_to_player = 0;

/*
 *
 * The rest of the read stuff, we do this in each forked child
 * This routine never returns it exits.
 *
 */
void manage_messages(to_mud, to_player, proc_name)
    int to_mud, to_player;
    char *proc_name;		/* We can change this to reflect player */
{
    struct timeval timeout;
    int res;
    fd_set readfds, writefds, exfds;
    int nfds;
    char namebuf[200], *playname, *t;

    nfds = 0;
    if (to_mud >= nfds)
	nfds = to_mud + 1;
    if (to_player >= nfds)
	nfds = to_player + 1;

    strcpy(namebuf,"M:           ");
    playname = &namebuf[2];
    left_to_mud = 0;
    left_to_player = 0;

    while(1)
    {
	FD_ZERO(&readfds);
	FD_ZERO(&writefds);
	FD_ZERO(&exfds);
	FD_SET(to_mud, &exfds);
	FD_SET(to_player, &exfds);

	if (left_to_mud && left_to_player)
	{
	    FD_SET(to_mud, &writefds);
	    FD_SET(to_player, &writefds);
	}
	else if (!left_to_mud && left_to_player)
	{
	    FD_SET(to_player, &readfds);
	    FD_SET(to_player, &writefds);
	}
	else if (left_to_mud && !left_to_player)
	{
	    FD_SET(to_mud, &readfds);
	    FD_SET(to_mud, &writefds);
	}
	else if (!left_to_mud && !left_to_player)
	{
	    FD_SET(to_mud, &readfds);
	    FD_SET(to_player, &readfds);
	}
	    
	timeout.tv_sec = 2; /* avoid busy waiting when no buffered cmds */
	timeout.tv_usec = 0;
	res = select(nfds, &readfds, &writefds, &exfds, &timeout);

	if (res == -1) 
	{
	    if (errno == EINTR) /* if we got an alarm, finish the round */
		continue;
	    perror("select");
	    exit(1);
	}
	
	if (FD_ISSET(to_player, &exfds) || FD_ISSET(to_mud, &exfds))
	    do_exit(to_player, to_mud);
	
	if (left_to_mud && FD_ISSET(to_mud, &writefds))
	{
	    t = send_msg(to_mud, left_to_mud, to_player);
	    free(left_to_mud);
	    left_to_mud = 0;
	    if (t)
		left_to_mud = t;
	}
	    
	if (left_to_player && FD_ISSET(to_player, &writefds))
	{
	    t = send_msg(to_player, left_to_player, to_mud);
	    free(left_to_player);
	    left_to_player = 0;
	    if (t)
		left_to_player = t;
	}

	if (!left_to_player && FD_ISSET(to_mud, &readfds))
	{
	    t = get_msg(to_mud, to_player);
	    if (t)
		left_to_player = string_copy(t);
	    else /* readbit set but nothing read => kill us */
		do_exit(to_mud, to_player);
	}
	    

	if (!left_to_mud && FD_ISSET(to_player, &readfds))
	{
	    t = get_msg(to_player, to_mud);
	    if (t)
		left_to_mud = string_copy(t);
	    else /* readbit set but nothing read => kill us */
		do_exit(to_mud, to_player);

	    /* Can we snoop the name of the player?
	    */
	    if (*playname && t)
	    {
		for (; (*t != 0 && *t != '\n' && playname < &namebuf[198] );
		     t++, playname++)
		    if (*t >= ' ')
			*playname = *t;
		    else
			playname--;

		if (*t == '\n' || playname >= &namebuf[198])
		{	
		    playname = &namebuf[198];
		    *playname = 0;
		    if (strlen(namebuf) > strlen(proc_name))
			strncpy(proc_name, namebuf, strlen(proc_name));
		    else
			strcpy(proc_name, namebuf);
		}
	    }
			
	}
    }
}

void do_exit(from , to)
    int from;
    int to;
{
    shutdown(to, 2);
    close(to);
    shutdown(from, 2);
    close(from);
    exit(1);
}
    
char *get_msg(from, to)
    int from;
    int to;
{
    int i, l, size;
    char eb[100], *p;
    static char buff[1024];

    size = 1000;

    if ((l = read(from, buff, size)) == -1) 
    {
	if (errno == ENETUNREACH) 
	{
	    log_file("connects","Net unreachable detected.\n");
	    do_exit(from, to);
	}
	if (errno == EHOSTUNREACH)
	{
	    log_file("connects","Host unreachable detected.\n");
	    do_exit(from, to);
	}
	if (errno == ETIMEDOUT) 
	{
	    log_file("connects","Connection timed out detected.\n");
	    do_exit(from, to);
	}
	if (errno == ECONNRESET) 
	{
	    log_file("connects","Connection reset by peer detected.\n");
	    do_exit(from, to);
	}
	if (errno == EWOULDBLOCK) 
	{
	    log_file("connects","read would block socket!\n");
	    do_exit(from, to);
	}
	if (errno == EMSGSIZE) 
	{
	    log_file("connects","read EMSGSIZE !\n");
	    return 0;
	}
	perror("read");
	sprintf(eb,"Unknown errno %d!\n", errno);
	log_file("connects", eb);
	do_exit(from, to);
    }
    if (l > 0)
    {
	buff[l] = 0;
	return buff;
    }
    else
    {
	return 0;
    }
}


char *send_msg(int to, char *buff, int from)
    int to;
    char *buff;
    int from;
{
    int l;
    
    l = write(to, buff, strlen(buff));
    if (l < 0)
	do_exit(to, from);
    if (l != strlen(buff))
	return string_copy(&buff[l]);
    else
	return 0;
}

