/*
   comm.c

   Setting up the communication part.

*/
#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <memory.h>
#ifndef sun
#include <fcntl.h>
#endif
#ifdef _AIX
#include <sys/select.h>
#endif

#include <sys/socket.h>
#ifdef _AIX
#include <sys/socketvar.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>

#define PROT(x) ()

int socket PROT((int, int, int));
#ifndef sgi
#ifdef NeXT
int setsockopt PROT((int, int, int, void *, int));
#else
int setsockopt PROT((int, int, int, char *, int));
#endif
int bind PROT((int, struct sockaddr *, int));
int listen PROT((int, int));
int accept PROT((int, struct sockaddr *, int *));
#endif /* sgi */
int select PROT((int, fd_set *, fd_set *, fd_set *, struct timeval *));

void bzero PROT((char *, int));

extern int errno;

static struct in_addr	ntohl_addr();

/*
 * Interprocess communication interface to the backend.
 */
/*
 * My data
 */
static int mud_s, udp_s, udp_send = -1;
extern int port_number;
extern int udp_port;
static struct sockaddr_in mudsin;
extern int send_hostname;

/*
 * The muds data
 */
extern int mud_port;
extern int mud_udp;
extern char *mud_host;

/* Prototypes
 */
void relay_udp();
int send_udp();
int mud_relay();
void relay_msg();
#ifndef INET_NTOA_OK
char inet_ntoa();
#endif
extern void log_file();
extern void print_error_file();

/*
 * Set up the defined ports
 */
void prepare_ipc() 
{
    struct sockaddr_in sin;
    struct hostent *hp;
    int tmp;
    char host_name[100], tbuf[1000];
    int n, ip1, ip2, ip3, ip4;

    /* First we set our server port
    */
    if (gethostname(host_name, sizeof host_name) == -1)
    {
        perror("gethostname");
	exit(1);
    }
    hp = gethostbyname(host_name);
    if (hp == 0) 
    {
	(void)fprintf(stderr, "gethostbyname: unknown host.\n");
	exit(1);
    }
    memset((char *)&sin, '\0', sizeof sin);
    memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);
    sin.sin_port = htons((u_short)port_number);
    sin.sin_family = hp->h_addrtype;
    sin.sin_addr.s_addr = INADDR_ANY;
    mud_s = socket(hp->h_addrtype, SOCK_STREAM, 0);
    if (mud_s == -1) 
    {
	perror("socket");
	abort();
    }
    tmp = 1;
    if (setsockopt (mud_s, SOL_SOCKET, SO_REUSEADDR,
		    (char *) &tmp, sizeof (tmp)) < 0) 
    {
	perror ("setsockopt");
	exit (1);
    }
    if (bind(mud_s, (struct sockaddr *)&sin, sizeof sin) == -1) 
    {
	if (errno == EADDRINUSE) 
	{
	    fprintf(stderr, "Socket already bound!\n");
	    exit(errno);
	} 
	else 
	{
	    perror("bind");
	    abort();
	}
    }
    if (listen(mud_s, 5) == -1) 
    {
	perror("listen");
	abort();
    }
    tmp = 1;

#if defined(sun) || defined(M_UNIX) || defined(NeXT)
    if (ioctl(mud_s, FIONBIO, &tmp) == -1) {
	perror("ioctl socket FIONBIO");
	abort();
    }
#else /* sun */
    if (fcntl(mud_s, F_SETFL, FNDELAY) == -1) {
	perror("ioctl socket FIONBIO");
	abort();
    }
#endif /* sun */

    signal(SIGPIPE, SIG_IGN);

    /* Then we set up our udp port
    */
    if (udp_port != -1)
    {
	sin.sin_port = htons((u_short)udp_port);
	udp_s = socket(hp->h_addrtype, SOCK_DGRAM, 0);
	if (udp_s == -1) 
	{
	    perror("udp_socket");
	    abort();
	}
	tmp = 1;
	if (setsockopt (udp_s, SOL_SOCKET, SO_REUSEADDR,
			(char *) &tmp, sizeof (tmp)) < 0) 
	{
	    perror ("setsockopt");
	    exit (1);
	}
	if (bind(udp_s, (struct sockaddr *)&sin, sizeof sin) == -1) 
	{
	    if (errno == EADDRINUSE) 
	    {
		fprintf(stderr, "UDP Socket already bound!\n");
		close(udp_s);
		udp_port = -1;
	    } 
	    else 
	    {
		perror("udp-bind");
		abort();
	    }
	}
    }
    if (udp_port != -1)
    {
	tmp = 1;

#if defined(sun) || defined(M_UNIX) || defined(NeXT)
	if (ioctl(udp_s, FIONBIO, &tmp) == -1) 
	{
	    perror("udp - ioctl socket FIONBIO");
	    abort();
	}
#else /* sun */
	if (fcntl(udp_s, F_SETFL, FNDELAY) == -1) 
	{
	    perror("udp - ioctl socket FIONBIO");
	    abort();
	}
#endif /* sun */
    }

    /* Then we set up the address to the mud
     */
    mudsin.sin_family = AF_INET;

    if (sscanf(mud_host,"%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4) == 4) 
    {
	mudsin.sin_addr.s_addr = inet_addr(mud_host);
    }
    else if((hp = gethostbyname(mud_host)) != NULL) 
    {
	bcopy(hp->h_addr, &mudsin.sin_addr, hp->h_length);
    }
    else 
    {
	fprintf(stderr,"gethostbyname: unknown mud host\n");
	abort();
    }
    mudsin.sin_port = htons(mud_port);
    sprintf(tbuf, "\nSetting up ipc: %d:%d(server), %d:%d(udp) %s:%d:%d(mud)\n",
	   port_number, mud_s, udp_port, udp_s, mud_host, mud_port, mud_udp);
    log_file("start", tbuf);
}

/*
 * This one is called when shutting down the Relay.
 * It is also called by the child to get rid of these sockets.
 */
void ipc_remove() 
{
    if (udp_send >= 0)
	close(udp_send);
    udp_send = -1;

    if (mud_s >= 0)
	close(mud_s);
    mud_s = -1;

    if (udp_s >= 0)
	close(udp_s);
    udp_s = -1;
}

/*
 * Stay in here until we have a new connection
 * This loop manages udp relay too.
 */
void new_connection(to_mud, to_player)
    int *to_mud;
    int *to_player;
{
    int new_socket, cnt, relay_socket;
    struct sockaddr_in addr;
    int length;
    struct timeval timeout;
    char udp_buf[1024], *st;
    fd_set readfds;
    int nfds, res;

    nfds = 0;
    if (mud_s >= nfds)
	nfds = mud_s + 1;
    if (udp_s >= nfds)
	nfds = udp_s + 1;

    while(1) 
    {
	FD_ZERO(&readfds);
	FD_SET(mud_s, &readfds);
	FD_SET(udp_s, &readfds);

	timeout.tv_sec = 2; /* avoid busy waiting */
	timeout.tv_usec = 0;
	res = select(nfds, &readfds, 0, 0, &timeout);

	if (res == -1) 
	{
	    if (errno == EINTR) /* if we got an alarm, finish the round */
		continue;
	    perror("select");
	    exit(1);
	}

	/* First, try to get a new connection...
	 */
	length = sizeof addr;
	new_socket = accept(mud_s, (struct sockaddr *)&addr, &length);
	if (new_socket != -1)
	{
	    relay_socket = mud_relay(&addr, length);
	    if (relay_socket >= 0)
	    {
		*to_mud = relay_socket;
		*to_player = new_socket;
		return;
	    }
	    else
	    {
		print_error_file(new_socket);
		shutdown(new_socket, 2);
		close(new_socket);
		new_socket = -1;
	    }
	}
	else if (new_socket == -1 && errno != EWOULDBLOCK && errno != EINTR)
	{
	    perror("accept");
	    abort();
	}

	/* Then see if we got any udp messages 
	*/
	if (udp_port != -1)
	{
	    cnt = recvfrom(udp_s, udp_buf, sizeof(udp_buf), 0, 
			   (struct sockaddr *)&addr, &length);
	    if (cnt != -1)
	    {
		udp_buf[sizeof(udp_buf) - 1] = 0;
		udp_buf[cnt] = 0;
		st = (char *)inet_ntoa(ntohl_addr(addr.sin_addr));
		relay_udp(st, udp_buf);
	    }
	}


	/* Whatever else might we want to do while we idle
	   We should do a global select to avoid busy wait.
	*/
    }
}

/*
 * Connect us to the mud and send the hostname of the caller to the mud
 * return the socket to the mud.
 */
int
mud_relay(addr, length)
    struct sockaddr_in *addr;
    int length;
{
    char buff[100], *p;
    struct hostent *hp;
    int s, l, i;

    if((s = socket(PF_INET, SOCK_STREAM, 0)) == -1)
    {
	perror("socket");
	return -1;
    }
  
    if(connect(s, &mudsin, sizeof(mudsin)) == -1) 
    {
	perror("connect");
	return -1;
    }
    if (!send_hostname)
	return s;

    hp = gethostbyaddr(&addr->sin_addr, length, AF_INET);

    /* Write the original callers hostname to the mud
    */
    if (hp)
	sprintf(buff, "%s\n", hp->h_name);
    else
	sprintf(buff, "%s\n", inet_ntoa(ntohl_addr(addr->sin_addr)));

    l = strlen(buff);
    
    for (p = buff; l > 0; l -= i)
    {
	i = write(s, p, l);
	if (i < l)
	    p = &p[i];
    }
    return s;
}

/*
 * This routine relays udp packages, wrapper:
 * From the mud:  RELAY(host:port)message
 * Sent to mud:   RELAY(host)message
 */
void
relay_udp(from_hostip, msg)
    char *from_hostip;
    char *msg;
{
    char mod_buf[2048], to_host[500], *rp, *lp, *colon;
    int to_port;
    
    /* Is it from the mud and on its way out ?
    */
    if (strcmp(from_hostip, mud_host) == 0)
    {
	/* this might just crash us if our mud sends a long faulty string
        */
	lp = strchr(msg, '(');
	rp = strchr(msg, ')');
	colon = strchr(msg, ':');

	if ((strlen(msg) > 5) && 
	    (strncmp(msg, "RELAY", 5) == 0) &&
	    lp && rp && colon)
	{
	    log_file("out_udp", msg);
	    *rp = 0; 
	    *colon = 0;
	    send_udp(&lp[1], atoi(&colon[1]), &rp[1]); 
	    *rp = ')';
	    *colon = ':';
	    return;
	}
    }

    if ((strlen(from_hostip) + strlen(msg) + 20) > 2047)
	return;
    sprintf(mod_buf, "RELAY(%s)%s", from_hostip, msg);
    log_file("in_udp", mod_buf);
    send_udp(mud_host, mud_udp, mod_buf);
}



/*
 * Send messages to other muds.
 *
 * The message is sent with udp. If it gets there it gets there...
 */
int send_udp(to_host, to_port, msg)
    char *to_host;
    int to_port;
    char *msg;
{
    int ip1, ip2, ip3, ip4;
    struct sockaddr_in name;
    struct hostent *hp, *gethostbyname();

    if (udp_send < 0)
    {
	/* Create socket on which to send. */
	udp_send = socket(AF_INET, SOCK_DGRAM, 0);
    }
	
    if (udp_send < 0)
	return 0;

    if (sscanf(to_host, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4) == 4) 
    {
	char *p = (char *)&name.sin_addr;
	p[0] = ip1;
	p[1] = ip2;
	p[2] = ip3;
	p[3] = ip4;
	name.sin_family = AF_INET;
	name.sin_addr = htonl(name.sin_addr);
    }
    else
    {
	hp = gethostbyname(to_host);
	if (hp == 0) {
	    fprintf(stderr, "%s: unknown host\n", to_host);
	    exit(2);
	}
	bcopy(hp->h_addr, &name.sin_addr, hp->h_length);
	name.sin_family = AF_INET;
    }
    name.sin_port = htons(to_port);
    /* Send message. */
    if (sendto(udp_send, msg, strlen(msg), 0, 
	       &name, sizeof(name)) != strlen(msg))
	return 0;
    return 1;
}    

/*
 * Some subfunctions
 */

static struct in_addr	ntohl_addr(addr)
    struct in_addr	addr;
{
	struct in_addr	ret;

	ret.s_addr = ntohl(addr.s_addr);
	return(ret);
}

/*
Note: if the address string is "a.b.c.d" the address number is
      a * 256^3 + b * 256^2 + c * 256 + d

*/
#ifndef INET_NTOA_OK

char *inet_ntoa(ad)
    struct in_addr ad;
{
    u_long s_ad;
    int a, b, c, d;
    static char addr[20]; /* 16 + 1 should be enough */

    s_ad = ad.s_addr;
    d = s_ad % 256;
    s_ad /= 256;
    c = s_ad % 256;
    s_ad /= 256;
    b = s_ad % 256;
    a = s_ad / 256;
    sprintf(addr, "%d.%d.%d.%d", a, b, c, d);
    return addr;
}

#endif
