/*
   main.c

   The main loop of the relayer.

*/
#include <signal.h>
#include <sys/wait.h>
#include <stdio.h>
#include <sys/time.h>

/*
 * Should be set by arguments on startup
 */
int port_number = 2000;
int udp_port = 2500;

char *mud_host = "129.16.125.130";
int mud_port = 3111;
int mud_udp = 2501;
int send_hostname = 1;
char *proc_name;
char *errfile;

/*
 * This should terminate children when people or the mud disconnects.
 * At least we give them a chance to do so
 */
void
do_wait()
{
    while (wait3(0, WNOHANG, 0) > 0)
	sleep(1);
}

/*
 * This should first shut down the server and udp socket then wait for
 * all children to terminate and then exit.
 */
void
do_hup()
{
    ipc_remove();
    strcpy(proc_name,"mudrelay_kill");
    while (wait() != -1)
	sleep(1);
    exit(0);
}


int main(argc, argv)
    int		argc;
    char	**argv;
{
    int to_mud, to_player, cpid, i;
    proc_name = argv[0];

    for (i = 1; i < argc; i++)
    {
	if (argv[i][0] != '-')
	{
	    fprintf(stderr, "Bad argument %s\n", argv[i]);
	    exit(1);
	}
	else 
	{
	    /*
	     * Look at flags.
	     *    -r     UDP port to receive on
	     *    -s     Server port, players connect here
	     *    -h     Mudhost to connect to
	     *    -m     Mudport the server port of the mud
             *    -u     Mududpport, where to send udp packages to the mud
             *    -n     No hostname sent to mud
	     *    -e     Name of errorfile to print if no mud to connect to
	     *    -?     Helptext and exit
	     */
	    switch(argv[i][1]) 
	    {
	    case 'r':
		udp_port = atoi(&argv[i][2]);
		continue;
	    case 's':
		port_number = atoi(&argv[i][2]);
		continue;
	    case 'h':
		mud_host = &argv[i][2];
		continue;
	    case 'm':
		mud_port = atoi(&argv[i][2]);
		continue;
	    case 'u':
		mud_udp = atoi (&argv[i][2]);
		continue;
	    case 'n':
		send_hostname = 0;
		continue;
	    case 'e':
		errfile = &argv[i][2];
		continue;
	    default:
		fprintf(stderr, "Unknown flag: %s\n", argv[i]);
		printf("Flags: \n\t-r     UDP port to receive on\n\t-s     Server port, players connect here\n\t-h     Mudhost to connect to\n\t-m     Mudport the server port of the mud\n\t-u     Mududpport, where to send udp packages to the mud\n\t-n     No hostname sent to mud\n\t-e     Name of errorfile to print when no mud\n\t-?     Helptext and exit\n");
		exit(1);
	    }
	}
    }
    
    prepare_ipc();

    signal(SIGCHLD, do_wait);
    signal(SIGHUP, do_hup);

    while (1)
    {
	new_connection(&to_mud, &to_player);
	if ((cpid = fork()) == 0)
	{
	    /* This never returns, it exits */
	    ipc_remove(); /* The child should not have server funct. */
	    signal(SIGCHLD, SIG_IGN);
	    signal(SIGHUP, SIG_IGN);
	    manage_messages(to_mud, to_player, argv[0]);
	}
	else /* The parent does not need those sockets */
	{
	    close(to_mud);
	    close(to_player);
	}
    }
    ipc_remove();
}



void
log_file(fname, msg)
    char *fname;
    char *msg;
{
    FILE *f;
    long t;

    if ((f = fopen(fname, "a+")) != NULL)
    {
	t = time(0l);
	fprintf(f, "%s-- %s\n", ctime(&t), msg);
	fclose(f);
    }
    else
	printf("Cant log\n");
}

char *string_copy(char *st)
{
    char *s;

    s = (char *)malloc(strlen(s) + 1);
    strcpy(s, st);
    return s;
}

void 
print_error_file(tofd)
    int tofd;
{
    FILE *f;
    char buff[2048], *p;
    int l, n;
    
    if ((f = fopen(errfile, "r")) != NULL)
    {
	n = 2000;
	l = fread(buff, &n, 1, f);
	buff[n + 1] = 0;

	for (l = 0, p = buff; l != strlen(p); p = &p[l])
	{
	    l = write(tofd, p, strlen(p));
	}
	fflush(tofd);
	fclose(f);
    }
}
	
